@@ -8,6 +8,7 @@ import subprocess
 import time
 from contextlib import contextmanager
 from datetime import datetime, timedelta
+from math import ceil
 from threading import Thread
 from typing import NamedTuple, Optional, Dict, List, Callable
 
@@ -16,18 +17,28 @@ from psycopg2._psycopg import DatabaseError
 from psycopg2.extras import NamedTupleCursor
 from psycopg2.pool import ThreadedConnectionPool
 
+from cortex_serving_client.deployment_failed import DeploymentFailed, COMPUTE_UNAVAILABLE_FAIL_TYPE, \
+    DEPLOYMENT_TIMEOUT_FAIL_TYPE, DEPLOYMENT_ERROR_FAIL_TYPE
 from cortex_serving_client.printable_chars import remove_non_printable
 
+
 """
 Details: https://www.cortex.dev/deployments/statuses
 On some places custom "not_deployed" status may be used.
 """
 
-CORTEX_STATUSES = ["live", "updating", "error", "error (out of memory)", "compute unavailable"]  #
+LIVE_STATUS = "live"
+UPDATING_STATUS = "updating"
+COMPUTE_UNAVAILABLE_STATUS = "compute unavailable"
+CORTEX_STATUSES = [LIVE_STATUS, UPDATING_STATUS, "error", "error (out of memory)", COMPUTE_UNAVAILABLE_STATUS]  #
+
 CORTEX_DELETE_TIMEOUT_SEC = 10 * 60
 CORTEX_DEPLOY_REPORTED_TIMEOUT_SEC = 60
 CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT = 20 * 60
 CORTEX_DEFAULT_API_TIMEOUT = CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT
+CORTEX_DEPLOY_RETRY_BASE_SLEEP_SEC = 5 * 60
+CORTEX_CMD_BASE_RETRY_SEC = 5
+CORTEX_STATUS_CHECK_SLEEP_SEC = 10
 INFINITE_TIMEOUT_SEC = 30 * 365 * 24 * 60 * 60  # 30 years
 CORTEX_DEFAULT_COMMAND_SYSTEM_PROCESS_TIMEOUT = 3 * 60
 
@@ -53,6 +64,7 @@ class CortexClient:
         deployment_timeout_sec=CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT,
         api_timeout_sec=CORTEX_DEFAULT_API_TIMEOUT,
         print_logs=False,
+        n_retries=0
     ) -> 'CortexGetResult':
         """
         Deploy an API until timeouts. Cortex docs https://docs.cortex.dev/deployments/deployment.
@@ -71,6 +83,8 @@ class CortexClient:
             Time until API will be auto-deleted. Use `INFINITE_TIMEOUT_SEC` for infinite.
         print_logs
             Subscribe to Cortex logs of the API and print them to stdout.
+        n_retries
+            Number of attempts to deploy until raising the failure. Retries only if the failure is not an application error.
 
         Returns
         -------
@@ -97,45 +111,65 @@ class CortexClient:
         # file has to be created the dir to which python predictors are described in yaml
         filename = f"{name}.yaml"
         filepath = f"{dir}/{filename}"
-        with str_to_public_temp_file(predictor_yaml_str, filepath):
-            self._collect_garbage()
-            gc_timeout_sec = api_timeout_sec + CORTEX_DELETE_TIMEOUT_SEC
-            logger.info(f"Deployment {name} has deployment timeout: {deployment_timeout_sec} sec, garbage collection timeout: {gc_timeout_sec} seconds.")
-            with self._open_db_cursor() as cursor:
-                cursor.execute(
-                    """
-                    insert into cortex_api_timeout(api_name, ultimate_timeout) values (%s, transaction_timestamp() + %s * interval '1 second')
-                    on conflict (api_name) do update
-                        set ultimate_timeout = transaction_timestamp() + %s * interval '1 second', modified = transaction_timestamp()
-                """,
-                    [name, gc_timeout_sec, gc_timeout_sec],
-                )
 
-            _verbose_command_wrapper(["cortex", "deploy", filename, f"--env={self.cortex_env}"], cwd=dir)
+        for retry in range(n_retries + 1):
+            try:
+                self._collect_garbage()
+                with str_to_public_temp_file(predictor_yaml_str, filepath):
+                    gc_timeout_sec = api_timeout_sec + CORTEX_DELETE_TIMEOUT_SEC
+                    logger.info(f"Deployment {name} has deployment timeout: {deployment_timeout_sec} sec, garbage collection timeout: {gc_timeout_sec} seconds.")
+                    with self._open_db_cursor() as cursor:
+                        cursor.execute(
+                            """
+                            insert into cortex_api_timeout(api_name, ultimate_timeout) values (%s, transaction_timestamp() + %s * interval '1 second')
+                            on conflict (api_name) do update
+                                set ultimate_timeout = transaction_timestamp() + %s * interval '1 second', modified = transaction_timestamp()
+                        """,
+                            [name, gc_timeout_sec, gc_timeout_sec],
+                        )
 
-        if print_logs:
-            self._cortex_logs_print_async(name)
+                    _verbose_command_wrapper(["cortex", "deploy", filename, f"--env={self.cortex_env}"], cwd=dir)
 
-        start_time = time.time()
-        while True:
-            get_result = self.get(name)
-            if get_result.status not in ("live", "updating"):
-                # avoids boot loop
-                self.delete(name)
-                raise ValueError(f"Deployment {name} failed with status {get_result.status}.")
+                if print_logs:
+                    self._cortex_logs_print_async(name)
 
-            elif get_result.status == "live":
-                break
+                start_time = time.time()
+                while True:
+                    get_result = self.get(name)
+                    time_since_start = ceil(time.time() - start_time)
+                    if get_result.status == LIVE_STATUS:
+                        return get_result
 
-            if start_time + deployment_timeout_sec < time.time():
-                # avoids boot loop
-                self.delete(name)
-                raise ValueError(f"Deployment {name} timeout after {deployment_timeout_sec} seconds.")
+                    elif get_result.status == COMPUTE_UNAVAILABLE_STATUS:
+                        self.delete(name)
+                        raise DeploymentFailed(
+                            f'Deployment of {name} API could not start due to insufficient memory, CPU, GPU or Inf in the cluster after {time_since_start} secs.',
+                            COMPUTE_UNAVAILABLE_FAIL_TYPE, name, time_since_start)
 
-            logger.info(f"Sleeping during deployment of {name} until next retry. Current get_result: {get_result}.")
-            time.sleep(10)
+                    elif get_result.status != UPDATING_STATUS:
+                        self.delete(name)
+                        raise DeploymentFailed(f"Deployment of {name} failed with status {get_result.status} after {time_since_start} secs.",
+                                               DEPLOYMENT_ERROR_FAIL_TYPE, name, time_since_start)
+
+                    if time_since_start > deployment_timeout_sec:
+                        self.delete(name)
+                        raise DeploymentFailed(f"Deployment of {name} timed out after {deployment_timeout_sec} secs.",
+                                               DEPLOYMENT_TIMEOUT_FAIL_TYPE, name, time_since_start)
+
+                    logger.info(f"Sleeping during deployment of {name} until next status check. Current get_result: {get_result}.")
+                    time.sleep(CORTEX_STATUS_CHECK_SLEEP_SEC)
+
+            except DeploymentFailed as e:
+                if retry == n_retries or e.failure_type == DEPLOYMENT_ERROR_FAIL_TYPE:
+                    raise e
+
+                else:
+                    sleep_secs = ceil(CORTEX_DEPLOY_RETRY_BASE_SLEEP_SEC * 2 ** retry)
+                    logger.warning(f'Retrying {retry + 1} time after sleep of {sleep_secs} secs due to deployment failure: {e}')
+                    time.sleep(sleep_secs)
+
+        raise RuntimeError('Execution should never reach here.')
 
-        return get_result
 
     @contextmanager
     def deploy_temporarily(
@@ -145,6 +179,7 @@ class CortexClient:
         deployment_timeout_sec=CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT,
         api_timeout_sec=CORTEX_DEFAULT_API_TIMEOUT,
         print_logs=False,
+        n_retries=0
     ) -> 'CortexGetResult':
         """
         Deploy an API until timeouts. Cortex docs https://docs.cortex.dev/deployments/deployment.
@@ -163,6 +198,8 @@ class CortexClient:
             Time until API will be auto-deleted. Use `INFINITE_TIMEOUT_SEC` for infinite.
         print_logs
             Subscribe to Cortex logs of the API and print them to stdout.
+        n_retries
+            Number of attempts to deploy until raising the failure. Retries only if the failure is not an application error.
 
         Returns
         -------
@@ -172,7 +209,7 @@ class CortexClient:
         """
 
         try:
-            yield self.deploy_single(deployment, dir, deployment_timeout_sec, api_timeout_sec, print_logs)
+            yield self.deploy_single(deployment, dir, deployment_timeout_sec, api_timeout_sec, print_logs, n_retries)
 
         finally:
             self.delete(deployment["name"])
@@ -209,7 +246,7 @@ class CortexClient:
     def _parse_get_deployed(cmd_out: str):
         lines = cmd_out.splitlines()
         status = lines[2].split("  ")[0].strip()
-        if status == "live":
+        if status == LIVE_STATUS:
             for line in lines:
                 line_split = line.split(" ")
                 if line_split[0] == "endpoint:":
@@ -230,24 +267,28 @@ class CortexClient:
             delete_arr.append("-f")
 
         accept_deletion_if_asked_input = b"y\n"
-        self._open_cursor_if_none(cursor, lambda c: self._del_db_api_row(name, c))
-        start_time = time.time()
-        while True:
-            get_result = self.get(name)
-            if get_result.status == "not_deployed":
-                return get_result
+        try:
+            self._open_cursor_if_none(cursor, self._modify_ultimate_timeout_to_delete_timeout, name)
+            start_time = time.time()
+            while True:
+                get_result = self.get(name)
+                if get_result.status == "not_deployed":
+                    return get_result
 
-            else:
-                _verbose_command_wrapper(delete_arr, input=accept_deletion_if_asked_input,
-                                         allow_non_0_return_code_on_stdout_sub_strs=['not deployed'])
+                else:
+                    _verbose_command_wrapper(delete_arr, input=accept_deletion_if_asked_input,
+                                             allow_non_0_return_code_on_stdout_sub_strs=['not deployed'])
 
-            if start_time + timeout_sec < time.time():
-                raise ValueError(
-                    f"Timeout after {timeout_sec} seconds. Attempted force delete, but not waiting for results."
-                )
+                if start_time + timeout_sec < time.time():
+                    raise ValueError(
+                        f"Timeout after {timeout_sec} seconds. Attempted force delete, but not waiting for results."
+                    )
 
-            logger.info(f"During delete of {name} sleeping until next retry. Current get_result: {get_result}.")
-            time.sleep(10)
+                logger.info(f"During delete of {name} sleeping until next status check. Current get_result: {get_result}.")
+                time.sleep(CORTEX_STATUS_CHECK_SLEEP_SEC)
+
+        finally:
+            self._open_cursor_if_none(cursor, self._del_db_api_row, name)
 
     def _cortex_logs_print_async(self, name):
         def listen_on_logs():
@@ -262,7 +303,21 @@ class CortexClient:
         worker.start()
 
     @staticmethod
-    def _del_db_api_row(name, cur):
+    def _modify_ultimate_timeout_to_delete_timeout(cur, name):
+        """ Will update modified time to prevent GC raise conditions and shorten timeout to the amount needed to delete the api."""
+        try:
+            cur.execute("""
+                update cortex_api_timeout
+                set ultimate_timeout = transaction_timestamp() + %s * interval '1 second',
+                    modified = transaction_timestamp()
+                where api_name = %s""",
+                [CORTEX_DELETE_TIMEOUT_SEC, name])
+
+        except DatabaseError:
+            logger.warning(f"Ignoring exception during cortex_api_timeout record for {name} modifying ultimate_timeout to delete the api", exc_info=True)
+
+    @staticmethod
+    def _del_db_api_row(cur, name):
         try:
             cur.execute("delete from cortex_api_timeout where api_name = %s", [name])
 
@@ -272,13 +327,13 @@ class CortexClient:
     def _open_db_cursor(self):
         return open_pg_cursor(self.db_connection_pool)
 
-    def _open_cursor_if_none(self, cursor, fun: Callable):
+    def _open_cursor_if_none(self, cursor, fun: Callable, *args):
         if cursor is not None:
-            fun(cursor)
+            fun(cursor, *args)
 
         else:
             with self._open_db_cursor() as cursor:
-                fun(cursor)
+                fun(cursor, *args)
 
     def _collect_garbage(self):
         logger.info(f"Starting garbage collection.")
@@ -296,7 +351,7 @@ class CortexClient:
 
                     else:
                         logger.warning(f"Collecting Cortex garbage - timed-out db row: {api_name}")
-                        self._del_db_api_row(api_name, cur)
+                        self._del_db_api_row(cur, api_name)
 
                 cur.execute("commit")
 
@@ -305,21 +360,19 @@ class CortexClient:
                 # Remove recorded but not deployed - fast to avoid conflicts
                 cur.execute(
                     "select api_name from cortex_api_timeout where modified + %s * interval '1 second' < transaction_timestamp() and not (api_name = ANY(%s))",
-                    [CORTEX_DEPLOY_REPORTED_TIMEOUT_SEC, deployed_api_names],
+                    [CORTEX_DELETE_TIMEOUT_SEC, deployed_api_names],
                 )
                 if cur.rowcount > 0:
                     apis = [r.api_name for r in cur.fetchall()]
                     logger.warning(f"Collecting Cortex garbage - recorded but not deployed: {apis}")
                     cur.execute(
                         "delete from cortex_api_timeout where modified + %s * interval '1 second' < transaction_timestamp() and not (api_name = ANY(%s))",
-                        [CORTEX_DEPLOY_REPORTED_TIMEOUT_SEC, deployed_api_names],
+                        [CORTEX_DELETE_TIMEOUT_SEC, deployed_api_names],
                     )
                     cur.execute("commit")
 
                 # Remove deployed but not recorded
-                cur.execute(
-                    "select api_name from cortex_api_timeout", [CORTEX_DEPLOY_REPORTED_TIMEOUT_SEC],
-                )
+                cur.execute("select api_name from cortex_api_timeout")
                 recorded_apis_set = set([r.api_name for r in cur.fetchall()])
                 for deployed_not_recorded_name in set(deployed_api_names).difference(recorded_apis_set):
                     logger.warning(f"Collecting Cortex garbage - deployed not recorded: {deployed_not_recorded_name}")
@@ -379,7 +432,7 @@ def _verbose_command_wrapper(
 ):
     cmd_str = " ".join(cmd_arr)
     message = ""
-    for retry in range(retry_count):
+    for retry in range(retry_count + 1):
         try:
             p = subprocess.run(cmd_arr, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd, timeout=timeout, input=input)
             stdout = p.stdout.decode()
@@ -393,21 +446,21 @@ def _verbose_command_wrapper(
                 logger.warning(f"Allowed unsuccessful command {cmd_arr} execution. Stdout {json.dumps(stdout)} or stderr {json.dumps(stderr)} matches one of {allow_non_0_return_code_on_stdout_sub_strs}")
                 return stdout
 
-            message = f"Non zero return code for command {cmd_str}! Stdout:\n{json.dumps(stdout)}"
-            if retry <= retry_count:
-                logger.info(message)
+            else:
+                message = f"Non zero return code for command {cmd_str}! Stdout:\n{json.dumps(stdout)}"
 
         except subprocess.TimeoutExpired as e:
-            timeout_message = f'{e} with stdout: "{e.output}" and stderr: "{e.stderr}"'
-            if retry <= retry_count:
-                logger.info(f"Retrying command {cmd_str} ignoring exception: " + timeout_message)
+            message = f'Timed out command  {cmd_str}: {e} with stdout: "{json.dumps(e.output)}" and stderr: "{json.dumps(e.stderr)}"'
 
-            else:
-                raise ValueError(f"Retry count for command {cmd_str} exceeded: " + timeout_message) from e
+        if retry < retry_count:
+            sleep_secs = ceil(CORTEX_CMD_BASE_RETRY_SEC * 2 ** retry)
+            logger.info(f"Retrying after {sleep_secs} sec: {message}")
+            time.sleep(sleep_secs)
 
-        time.sleep(3)
+        else:
+            raise ValueError(f"Retry count for command {cmd_str} exceeded: {message}")
 
-    raise ValueError(f"Retry count for command {cmd_str} exceeded: " + message)
+    raise RuntimeError(f'Execution should never reach here: {message}')
 
 
 class CortexGetAllStatus(NamedTuple):
